// made by: Scott Klasky 1998
// This is a 3D elliptic equation $ u_{xx} + u_{yy} + u_{zz} = f(x,y,z) $
// With Dirichlet boundaries, in this "model" problem, we let the exact solution be
// u = 1/sqrt(x^2+y^2+z^2+1/2), so we can find f from this.
// Since elliptic solvers depend on L .... [ Lu = f], and not on f,
// we can pose a problem like this.
// The output will contain 2 columns, the work, and the l1 norm of the residual. 
// Notice that the residuals go down by about a factor of 10, for ever v-cycle.

import java.io.*;

public class mg3d{

  public int power(int number,int pow){
    int sum = 1;
    for (int i=0;i<pow;i++) {
      sum = sum * number;
    }
    return sum;
  }

  static void print_uij(int nx,int ny, int nz, double x[], double y[],
			double z[], double val[][][],String name)
    throws java.io.IOException{
      PrintStream out;
      out = new PrintStream( new FileOutputStream(name));
      int j = (ny-1)/2;
      int k = (nz-1)/2;
      for (int i=0;i<nx;i++) {
	out.println(x[i]+" "+val[k][j][i]);
      }
  }

  static void set_up_u(int nx, int ny, int nz, double x[],
		       double y[],double z[],
		       double u[][][], double uxct[][][], double h){
    for (int k=0;k<nz;k++) {
      for (int j=0;j<ny;j++) {
	for (int i=0;i<nx;i++) {
	  uxct[k][j][i] = 1.0/Math.sqrt(x[i]*x[i]+y[j]*y[j]+z[k]*z[k]+0.5);
	  u[k][j][i] = 0.0;
	}
      }
    }
    for (int k=0;k<nz;k++) {
      for (int j=0;j<ny;j++) {
	u[k][j][0] = uxct[k][j][0];
	u[k][j][nx-1] = uxct[k][j][nx-1];
      }
    }
    for (int k=0;k<nz;k++) {
      for (int i=0;i<nx;i++) {
	u[k][0][i] = uxct[k][0][i];
	u[k][ny-1][i] = uxct[k][ny-1][i];
      }
    }
    for (int j=0;j<ny;j++) {
      for (int i=0;i<nx;i++) {
	u[0][j][i] = uxct[0][j][i];
	u[nz-1][j][i] = uxct[nz-1][j][i];
      }
    }

  }


  static void set_up_rhs(int nx,int ny,int nz,double x[],double y[],double z[],
			 double rhs[][][]){
    for (int k=0;k<nz;k++) {
      for (int j=0;j<ny;j++) {
	for (int i=0;i<nx;i++) {
	  rhs[k][j][i] = -48/Math.sqrt(
				       (4*x[i]*x[i]+4*y[j]*y[j]+4*z[k]*z[k]+2.0)*
				       (4*x[i]*x[i]+4*y[j]*y[j]+4*z[k]*z[k]+2.0)*
				       (4*x[i]*x[i]+4*y[j]*y[j]+4*z[k]*z[k]+2.0)*
				       (4*x[i]*x[i]+4*y[j]*y[j]+4*z[k]*z[k]+2.0)*
				       (4*x[i]*x[i]+4*y[j]*y[j]+4*z[k]*z[k]+2.0));
	}
      }
    }
  }

  static void restrict(int nxf, int nyf, int nzf, double uf[][][], 
		       double uc[][][] ){
    double HALF = 0.5;
    double TWELTH = 1.0/12.0;
    int nxc = (nxf+1)/2;
    int nyc = (nyf+1)/2;
    int nzc = (nzf+1)/2;


    for (int kc=1; kc<nzc-1; kc ++) {
      int kff = 2 * kc;
      for (int jc=1; jc<nyc-1; jc ++) {
	int jff = 2 * jc;
	for (int ic=1; ic<nxc-1; ic ++) {
	  int iff = 2 * ic;
	  uc[kc][jc][ic] = HALF*uf[kff][jff][iff] + TWELTH*(
							    uf[kff+1][jff][iff] + uf[kff-1][jff][iff] +
							    uf[kff][jff+1][iff] + uf[kff][jff-1][iff] +
							    uf[kff][jff][iff+1] + uf[kff][jff][iff-1] );
	}
      }
    }
    // inject the boundaries....
    for (int jc=0;jc<nyc;jc++) {
      for (int ic=0;ic<nxc;ic++) {
	uc[0][jc][ic] = uf[0][2*jc][2*ic];
	uc[nzc-1][jc][ic] = uf[nzf-1][2*jc][2*ic];
      }
    }
    for (int kc=0;kc<nyc;kc++) {
      for (int ic=0;ic<nxc;ic++) {
	uc[kc][0][ic] = uf[2*kc][0][2*ic];
	uc[kc][nyc-1][ic] = uf[2*kc][nyf-1][2*ic];
      }
    }
    for (int kc=0;kc<nyc;kc++) {
      for (int jc=0;jc<nyc;jc++) {
	uc[kc][jc][0] = uf[2*kc][2*jc][0];
	uc[kc][jc][nxc-1] = uf[2*kc][2*jc][nxf-1];
      }
    }

  }


  static void prolong(int nxf, int nyf, int nzf, 
		      double uc[][][], double uf[][][] ){
    int nxc = (nxf+1)/2;
    int nyc = (nyf+1)/2;
    int nzc = (nzf+1)/2;
    int iff,jff,kff;
    double HALF = 0.5;


    for (int kc=0; kc<nzc;kc ++) {
      kff = 2 * kc;
      for (int jc=0; jc<nyc; jc ++) {
	jff = 2 * jc;
	for (int ic=0; ic<nxc; ic ++) {
	  iff = 2 * ic;
	  uf[kff][jff][iff] =uc[kc][jc][ic];   // injection
	}
	for (iff=1;iff<nxf-1;iff+=2) {
	  uf[kff][jff][iff]=HALF*(uf[kff][jff][iff-1]+uf[kff][jff][iff+1]);
	}
	if (jc > 0 ) {
	  for (iff=0;iff<nxf;iff++) {
	    uf[kff][jff-1][iff] = HALF*(uf[kff][jff-2][iff] + 
					uf[kff][jff][iff] );
	  }
	}
      }
      if (kc > 0 ) {
	for (jff=0;jff<nyf;jff++) {
	  for (iff=0;iff<nxf;iff++) {
	    uf[kff-1][jff][iff] = HALF*(uf[kff-2][jff][iff] +
					uf[kff][jff][iff] );
	  }
	}
      }
    }
  }

  static void lu(int nx, int ny, int nz, double u[][][],
		 double h, double lu[][][] ){

    double h2    = 1.0/( h * h);
    double jaceli=  1.0/(6.0 * (h * h));
    double sum = 0.0;

    for (int k=0;k<nz;k++) {
      for (int j=0;j<ny;j++) {
	for (int i=0;i<nx;i++) {
	  lu[k][j][i] = 0.0;
	}
      }
    }

    for (int k=1;k<nz-1;k++) {
      for (int j=1;j<ny-1;j++) {
	for (int i=1;i<nx-1;i++) {

	  lu[k][j][i] = h2 * ( u[k][j+1][i] + u[k][j-1][i] +
                               u[k][j][i+1] + u[k][j][i-1] +
                               u[k+1][j][i] + u[k-1][j][i] -
                               6.0 * u[k][j][i] );
	}
      }
    }
  }


  static void suba(int nx, int ny, int nz, double u1[][][],
		   double u2[][][],double u3[][][]){
    for (int k=0;k<nz;k++) {
      for (int j=0;j<ny;j++) {
	for (int i=0;i<nx;i++) {
	  u3[k][j][i] = u1[k][j][i] - u2[k][j][i];
	}
      }
    }
  }

  static void adda(int nx, int ny, int nz, double u1[][][],
		   double u2[][][],double u3[][][]){
    for (int k=0;k<nz;k++) {
      for (int j=0;j<ny;j++) {
	for (int i=0;i<nx;i++) {
	  u3[k][j][i] = u1[k][j][i] + u2[k][j][i];
	}
      }
    }
  }


  double erroru(int nx, int ny, int nz, double u[][][], double ux[][][]){
    double error;
    double val;
    error = 0.0;
    for (int k=0;k<nz;k++){
      for (int j=0;j<ny;j++){
	for (int i=0;i<nx;i++){
	  val = u[k][j][i]-ux[k][j][i];
	  error = error + val * val;
	}
      }
    }
    error = error / nx /ny/nz;
    return error;
  }


  double relaxc(int nx, int ny, int nz, double u[][][],double rhs[][][],
		double res[][][], double h){

    double h2    = 1.0/( h * h);
    double jacel = (h * h)/ 6.0;
    double sum = 0.0;

    int ksw = 0;
    int isw;
    for (int pass=1;pass<3;pass++){
      isw = ksw;
      for (int k=1;k<nz-1;k++) {
	for (int j=1;j<ny-1;j++) {
	  for (int i=isw+1;i<nx-1;i+=2) {

	    res[k][j][i] = h2 * ( u[k][j+1][i] + u[k][j-1][i] +
                                  u[k][j][i+1] + u[k][j][i-1] +
                                  u[k+1][j][i] + u[k-1][j][i] -
                                  6.0 * u[k][j][i] )  - rhs[k][j][i];
   
	    u[k][j][i] += jacel * res[k][j][i];
	    sum = sum + res[k][j][i] * res[k][j][i] ;
	  }
	  isw=1 - isw;
	}
      }
      ksw = 1 - ksw;
    }
    double sum2 = sum / nx / nx / ny / ny / nz / nz;
    double sum3 = Math.sqrt(sum2);

    return sum3;
  }


  public static void main(String[] args)throws java.io.IOException{
    mg3d relax = new mg3d();
    mg3d err   = new mg3d();
    double WORK = 0.0, WORK_NORM=0.0;
    double EPSILON = 0.0000000001, EPS=0.0;
    double alpha = 1.0/3.0;
    if (args.length<4) {
      System.err.println(
			 "Usage: java mg3d lev[nx=2^level + 1]  ncycles[# of v cycles]  pre [ # of pre smoothing sweeps, typically 3 pst [# of post smoothing sweeps, typically 3");
      return;
    }
    mg3d mm = new mg3d();
    // now initialize the memory.....
    int levels,maxlev;
    levels = Integer.parseInt(args[0]);
    int ncycle = Integer.parseInt(args[1]);
    int pre = Integer.parseInt(args[2]);
    int pst = Integer.parseInt(args[3]);

    int[] nx = new int[levels];
    int[] ny = new int[levels];
    int[] nz = new int[levels];
    int nxf, nyf, nzf;
    // use this in the c type loops, or should I say java loops
    maxlev = levels - 1;

    // declare the 2d variables on all of the levels ....
    double [][][][] u = new double[levels][][][];
    double [][][][] uxct = new double[levels][][][];
    double [][][][] rhs  = new double[levels][][][];
    double [][][][] tau  = new double[levels][][][];
    double [][][][] temp  = new double[levels][][][];
    //
      
    for (int i=0;i<levels;i++) {
      nx[i] = 1 + mm.power(2,i+1); 
      ny[i] = 1 + mm.power(2,i+1); 
      nz[i] = 1 + mm.power(2,i+1); 

      u[i]    = new double[nz[i]][ny[i]][nx[i]];
      uxct[i] = new double[nz[i]][ny[i]][nx[i]];
      rhs[i]  = new double[nz[i]][ny[i]][nx[i]];
      tau[i]  = new double[nz[i]][ny[i]][nx[i]];
      temp[i]  = new double[nz[i]][ny[i]][nx[i]];
    }

    nxf = nx[maxlev];
    nyf = ny[maxlev];
    nzf = nz[maxlev];
    WORK_NORM = 1.0 / (nxf*nyf*nzf);
    System.out.println("nx,ny,nz on finest grid= "+nxf+" "+nyf+" "+nzf);
    System.out.println("work                l1_norm residuals");
    double x[] = new double[nxf];
    double y[] = new double[nyf];
    double z[] = new double[nzf];
    double h;
    double l2norm=0.0,l2err=0.0;
    h = (2.0)/(nxf-1);

    for (int i=0;i<nxf;i++) {
      x[i] = -1.0 + (i)*h;
    }
    for (int i=0;i<nyf;i++) {
      y[i] = -1.0 + (i)*h;
    }
    for (int i=0;i<nzf;i++) {
      z[i] = -1.0 + (i)*h;
    }


    // put the data down on the finest level.....

    set_up_u(nxf,nyf,nzf,x,y,z,u[maxlev],uxct[maxlev],h);
    set_up_rhs(nxf,nyf,nzf,x,y,z,rhs[maxlev]); 

    // first smooth on the finest grid
    for (int prec =0;prec<pre;prec++) {
      l2norm = relax.relaxc(nx[maxlev],ny[maxlev],nz[maxlev],u[maxlev],
			    rhs[maxlev],temp[maxlev],h);
      WORK = WORK + nx[maxlev]*ny[maxlev]*nz[maxlev] * WORK_NORM;
    }
    //    l2err = err.erroru(nx[maxlev],ny[maxlev],nz[maxlev],u[maxlev],
    //	       uxct[maxlev]);
    System.out.println(+ WORK +"   " + l2norm);
    EPS = alpha * l2norm;
    // loop over a v cycle
      
    for (int cycle=0;cycle<ncycle;cycle++) {

      // cycle from the finest to the coarsest
      for (int lev=maxlev;lev>=1;lev--) {
	lu(nx[lev],ny[lev],nz[lev],u[lev],h,temp[lev]);
	restrict(nx[lev],ny[lev],nz[lev],temp[lev],temp[lev-1]);
	restrict(nx[lev],ny[lev],nz[lev],u[lev],u[lev-1]);
	h = h * 2.0;
	lu(nx[lev-1],ny[lev-1],nz[lev-1],u[lev-1],h,tau[lev-1]);
	suba(nx[lev-1],ny[lev-1],nz[lev-1],tau[lev-1],
	     temp[lev-1],tau[lev-1]);
	restrict(nx[lev],ny[lev],nz[lev],rhs[lev],rhs[lev-1]);
	adda(nx[lev-1],ny[lev-1],nz[lev-1],
	     rhs[lev-1],tau[lev-1],rhs[lev-1]);
	// at this point the truncation error has been added to the rhs 
	// now we need to smooth on this grid (or solve.....)
	if ( (lev-1) == 0) {
	  // solve on the grid.....
	  do{
	    l2norm = relax.relaxc(nx[lev-1],ny[lev-1],nz[lev-1],u[lev-1],
				  rhs[lev-1],tau[lev-1],h);
	    WORK = WORK + nx[lev-1]*ny[lev-1]*nz[lev-1] * WORK_NORM;
	    //l2err = err.erroru(nx[lev-1],ny[lev-1],nz[lev-1],u[lev-1],uxct[lev-1]);
	  } while (l2norm > EPS);

	}
	else {
	  for (int prec =0;prec<pre;prec++) {
	    l2norm = relax.relaxc(nx[lev-1],ny[lev-1],nz[lev-1],u[lev-1],
				  rhs[lev-1],temp[lev-1],h);
	    WORK = WORK + nx[lev-1]*ny[lev-1]*nz[lev-1] * WORK_NORM;
	  }
	  EPS = alpha * l2norm;
	}

      }

      // now cycle back up to the finest grid (but not including the finest level
      for (int lev=0;lev<maxlev;lev++) {
	restrict(nx[lev+1],ny[lev+1],nz[lev+1],u[lev+1],temp[lev]);
	suba(nx[lev],ny[lev],nz[lev],u[lev],temp[lev],u[lev]);
	prolong(nx[lev+1],ny[lev+1],nz[lev+1],u[lev],temp[lev+1]);
	adda(nx[lev+1],ny[lev+1],nz[lev+1],u[lev+1],temp[lev+1],u[lev+1]);
	h = h * 0.5;
	// now we apply some more smoothing sweeps
	for (int prec =0;prec<pre;prec++) {
	  l2norm = relax.relaxc(nx[lev+1],ny[lev+1],nz[lev+1],u[lev+1],
				rhs[lev+1],temp[lev+1],h);
	  WORK = WORK + nx[lev+1]*ny[lev+1]*nz[lev+1] * WORK_NORM;
	}
	EPS = alpha * l2norm;
	if (lev+1 == maxlev) {
	  System.out.println(+ WORK +"   " + l2norm);
	}
      }
      
    }  // the end of a vcycle
    // now we just have swept on the finest grid!!!


    // now print out the data.....

    print_uij(nx[maxlev],ny[maxlev],nz[maxlev],x,y,z,u[maxlev],"u.dat");
    print_uij(nx[maxlev],ny[maxlev],nz[maxlev],x,y,z,uxct[maxlev],"ux.dat");
    suba(nx[maxlev],ny[maxlev],nz[maxlev],u[maxlev],
	 uxct[maxlev],temp[maxlev]);
    print_uij(nx[maxlev],ny[maxlev],nz[maxlev],x,y,z,temp[maxlev],"err.dat");

  }
}
